/*
 * @(#)SimpleRoundedButtonUI.java
 *
 * Copyright (c) 2008 Jeremy Wood. All Rights Reserved.
 *
 * You have a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) You do not utilize the software in a manner
 * which is disparaging to the original author.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. THE AUTHOR SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL THE
 * AUTHOR BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */

package com.bric.plaf;

import java.util.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.MouseInputAdapter;

import java.awt.event.*;
import java.awt.*;
import java.awt.geom.*;

import javax.swing.plaf.basic.*;

/** A simple, textured-looking button UI.
 *
 * <P>Similar to how Apple customizes buttons, this UI
 * looks for the client property "JButton.segmentPosition".
 * How this UI renders will depend on whether this property is
 * "first", "last", "middle", or "only".
 * (It is assumed there will be no padding between two buttons
 * if they have adjacent button types.)
 */
public class SimpleRoundedButtonUI extends BasicButtonUI {
	private static final int LEFT = 0;
	private static final int RIGHT = 1;
	private static final int MIDDLE = 2;
	private static final int ONLY = 3;

	private boolean pressed = false;
	private MouseListener myMouseListener = new MouseAdapter() {
		public void mousePressed(MouseEvent e) {
			pressed = true;
		}
		public void mouseReleased(MouseEvent e) {
			pressed = false;
		}
	};
	
	private static int getPosition(JComponent button) {
		String s = (String)button.getClientProperty("JButton.segmentPosition");
		
		if(s!=null && s.equals("first"))
			return LEFT;
		if(s!=null && s.equals("last"))
			return RIGHT;
		if(s!=null && s.equals("middle"))
			return MIDDLE;
		
		return ONLY;
	}
	
	private static Color focusColor = getFocusColor();
	
	private static Color getFocusColor() {
		Color c = UIManager.getColor("Focus.color");
		if(c!=null) return c;
		c = UIManager.getColor("Button.focus");
		c = new Color(c.getRed(),c.getGreen(),c.getBlue(),60);
		if(c!=null) return c;
		return new Color(40,40,40);
	}
	

	public void paint(Graphics g0, JComponent c) {
		Graphics2D g = (Graphics2D)g0;
		
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
				RenderingHints.VALUE_ANTIALIAS_ON );
		Shape shape = getShape(c,true);

		boolean selected = false;
		if(c instanceof AbstractButton) {
			selected = ((AbstractButton)c).isSelected();
		}

		g.translate(0, 1);
		g.setColor(new Color(255,255,255,100));
		g.draw(shape);
		g.translate(0, -1);
		shape = getShape(c,false);
		
		//these colors appear to most resemble the buttons in Mail:
		//g.setPaint(getBackground(c.getHeight(),213,200,192,216));
		
		//but I need something lighter here:
		ButtonModel model = null;
		if(c instanceof AbstractButton) {
			model = ((AbstractButton)c).getModel();
		}
		if(pressed || selected || (model!=null && model.isArmed())) {
			g.setPaint(getBackground(c.getHeight(),200,190,200,210));
		} else {
			g.setPaint(getBackground(c.getHeight(),240,230,220,230));
		}
		g.fill(shape);
		
		if(c.hasFocus()) {
			Shape oldClip = g.getClip();
			g.clip(shape);
			g.setStroke(new BasicStroke(4));
			g.setColor(focusColor);
			g.draw(shape);
			g.setClip(oldClip);
		}

		g.setColor(Color.gray);
		g.setStroke(new BasicStroke(1));
		g.draw(shape);

		if(c.isEnabled()==false) {
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,.5f));
		}
		super.paint(g,c);
	}



	protected void paintFocus(Graphics g, AbstractButton b, Rectangle viewRect,
			Rectangle textRect, Rectangle iconRect) {
		super.paintFocus(g, b, viewRect, textRect, iconRect);
	}



	public void installUI(JComponent c) {
		super.installUI(c);
		c.addMouseListener(myMouseListener);
		c.setOpaque(false);
	}



	public void uninstallUI(JComponent c) {
		super.uninstallUI(c);
		c.removeMouseListener(myMouseListener);
	}



	/** In keeping with convention in BasicButtonUI, let's recycle
	 * static variables.  This method will only be called from the
	 * AWT thread, so we're safe.
	 */
	private static Insets insets = new Insets(2,2,2,2);
	private static AffineTransform translation = new AffineTransform();
	private static GeneralPath path = new GeneralPath();
	
	private static Shape getShape(JComponent component,boolean lowerHalfOnly) {
		path.reset();
		int myPosition = getPosition(component);
		int x = insets.left;
		int y = insets.right;
		int w = component.getWidth()-insets.right-insets.left;
		int h = component.getHeight()-insets.top-insets.bottom;
		
		int r = Math.min(w+x, h+y)/2;
		
		if(myPosition==MIDDLE) {
			x = 0;
			w = w+insets.left+insets.right;
		} else if(myPosition==LEFT) {
			w = w+insets.right;
		} else if(myPosition==RIGHT) {
			x = 0;
			w = w+insets.left;
		}

		path.moveTo(w, h-r);
		if(myPosition==RIGHT || myPosition==ONLY) {
			path.curveTo(w, h, 
					w, h, 
					w-r,h);
		} else {
			path.lineTo(w, h);
			
		}
		
		if(myPosition==LEFT || myPosition==ONLY) {
			path.lineTo(r, h);
			path.curveTo(0, h, 
					0, h, 
					0,h-r);
		} else {
			path.lineTo(0,h);
		}
		
		if(lowerHalfOnly==false) {
			if(myPosition==LEFT || myPosition==ONLY) {
				path.lineTo(0, r);
				path.curveTo(0, 0, 
						0, 0, 
						r,0);
			} else {
				path.lineTo(0,0);
			}
			if(myPosition==RIGHT || myPosition==ONLY) {
				path.lineTo(w-r,0);
				path.curveTo(w, 0, 
						w, 0, 
						w,r);
				path.lineTo(w, h-r);
			} else {
				path.lineTo(w,0);
			}
			path.closePath();
		}
		
		translation.setToTranslation(x, y);
		path.transform(translation);
		
		return path;
	}
	
	private static Hashtable backgrounds = new Hashtable();
	private static Paint getBackground(int h,int g1,int g2,int g3,int g4) {
		String key = h+" "+g1+" "+g2+" "+g3+" "+g4;
		Paint paint = (Paint)backgrounds.get(key);
		if(paint==null) {
			BufferedImage bi = new BufferedImage(1,h,BufferedImage.TYPE_INT_RGB);
			int[] data = new int[h];
			for(int a = 0; a<data.length; a++) {
				float y = ((float)a)/((float)h);
				int g;
				if(y<.5f) {
					y = 2*y;
					g = (int)(g1*(1-y)+g2*y);
				} else {
					y = (y-.5f)*2;
					g = (int)(g3*(1-y)+g4*y);
				}
				data[a] = (0xff000000)+(g << 16)+(g<<8)+g;
			}
			bi.getRaster().setDataElements(0,0,1,h,data);
			paint = new TexturePaint(bi,new Rectangle(0,0,1,h));
			backgrounds.put(key,paint);
		}
		return paint;
	}

	/** @return the color used to indicate when a component has
	 * focus.  By default this uses the color (64,113,167), but you can
	 * override this by calling:
	 * <BR><code>UIManager.put("focusRing",customColor);</code>
	 */
	public static Color getFocusRingColor() {
		Object obj = UIManager.getColor("focusRing");
		if(obj instanceof Color)
			return (Color)obj;
		return new Color(64,113,167);
	}
	
	/** Paints 3 different strokes around a shape to indicate focus.
	 * The widest stroke is the most transparent, so this achieves a nice
	 * "glow" effect.
	 * <P>The catch is that you have to render this underneath the shape,
	 * and the shape should be filled completely.
	 * 
	 * @param g the graphics to paint to
	 * @param shape the shape to outline
	 * @param biggestStroke the widest stroke to use.
	 */
	public static void paintFocus(Graphics2D g,Shape shape,int biggestStroke) {
		Color focusColor = getFocusRingColor();
		Color[] focusArray = new Color[] {
			new Color(focusColor.getRed(), focusColor.getGreen(), focusColor.getBlue(),255),
			new Color(focusColor.getRed(), focusColor.getGreen(), focusColor.getBlue(),170),
			new Color(focusColor.getRed(), focusColor.getGreen(), focusColor.getBlue(),110)	
		};
		g.setStroke(new BasicStroke(biggestStroke));
		g.setColor(focusArray[2]);
		g.draw(shape);
		g.setStroke(new BasicStroke(biggestStroke-1));
		g.setColor(focusArray[1]);
		g.draw(shape);
		g.setStroke(new BasicStroke(biggestStroke-2));
		g.setColor(focusArray[0]);
		g.draw(shape);
		g.setStroke(new BasicStroke(1));
	}
	

    /** Returns the bounds of a given shape.
     * <P>This is more accurate than <code>GeneralPath.getBounds()</code>
     * and more efficient than <code>Area.getBounds()</code>, and it is
     * abstract enough to work on any type of shape.
     * 
     * @param i the path data you want the bounds of
     * @param dest if non-null, then the bounds will be stored in this rectangle.
     * @return a rectangle indicating the exact bounds of this shape.
     * The return value will be <code>dest</code> if <code>dest</code> is
     * not null.
     * @see <a href="http://javagraphics.blogspot.com/2007/05/shapes-calculating-bounds.html">shapes-calculating-bounds</A>
     * @throws RuntimeException if the path iterator contains no shape data.
     */
    public static Rectangle2D getBounds(PathIterator i,Rectangle2D dest) {
        if(dest==null)
            dest = new Rectangle2D.Double();
        
        double[] f = new double[6];
        int k;
        
        /** left, top, right, and bottom bounds, respectively */
        double[] bounds = new double[4];
        
        double lastX = 0;
        double lastY = 0;
        
        //A, B, C, and D in the polynomial equation x = a*t^3+b*t^2+c*t+d
        //or A, B, and C in the equation x = a*t^2+b*t+c
        double ax, bx, cx, dx, ay, by, cy, dy;
        
        double t, x, y, det;
        boolean defined = false;
        while(i.isDone()==false) {
            k = i.currentSegment(f);
            if(k==PathIterator.SEG_MOVETO) {
                lastX = f[0];
                lastY = f[1];
            } else if(k==PathIterator.SEG_CLOSE) {
                //do nothing
                //note if we had a simple MOVETO and SEG_CLOSE then
                //we haven't changed "bounds".  This is intentional,
                //so if the shape is badly defined the bounds
                //should still make sense.
            } else {
                if(defined == false) {
                    bounds[0] = lastX;
                    bounds[1] = lastY;
                    bounds[2] = lastX;
                    bounds[3] = lastY;
                    defined = true;
                } else {
                    if(lastX<bounds[0]) bounds[0] = lastX;
                    if(lastY<bounds[1]) bounds[1] = lastY;
                    if(lastX>bounds[2]) bounds[2] = lastX;
                    if(lastY>bounds[3]) bounds[3] = lastY;
                }
                
                if(k==PathIterator.SEG_LINETO) {
                    if(f[0]<bounds[0]) bounds[0] = f[0];
                    if(f[1]<bounds[1]) bounds[1] = f[1];
                    if(f[0]>bounds[2]) bounds[2] = f[0];
                    if(f[1]>bounds[3]) bounds[3] = f[1];
                    lastX = f[0];
                    lastY = f[1];
                } else if(k==PathIterator.SEG_QUADTO) {
                    //check the end point
                    if(f[2]<bounds[0]) bounds[0] = f[2];
                    if(f[3]<bounds[1]) bounds[1] = f[3];
                    if(f[2]>bounds[2]) bounds[2] = f[2];
                    if(f[3]>bounds[3]) bounds[3] = f[3];
                    
                    //find the extrema
                    ax = lastX-2*f[0]+f[2];
                    bx = -2*lastX+2*f[0];
                    cx = lastX;
                    ay = lastY-2*f[1]+f[3];
                    by = -2*lastY+2*f[1];
                    cy = lastY;
                    
                    //x = a*t^2+b*t+c
                    //dx/dt = 0 = 2*a*t+b
                    //t = -b/(2a)
                    t = -bx/(2*ax);
                    if(t>0 && t<1) {
                        x = (ax*t+bx)*t+cx;
                        if(x<bounds[0]) bounds[0] = x;
                        if(x>bounds[2]) bounds[2] = x;
                    }
                    t = -by/(2*ay);
                    if(t>0 && t<1) {
                        y = (ay*t+by)*t+cy;
                        if(y<bounds[1]) bounds[1] = y;
                        if(y>bounds[3]) bounds[3] = y;
                    }
                    lastX = f[2];
                    lastY = f[3];
                } else if(k==PathIterator.SEG_CUBICTO) {
                    if(f[4]<bounds[0]) bounds[0] = f[4];
                    if(f[5]<bounds[1]) bounds[1] = f[5];
                    if(f[4]>bounds[2]) bounds[2] = f[4];
                    if(f[5]>bounds[3]) bounds[3] = f[5];
                    
                    ax = -lastX+3*f[0]-3*f[2]+f[4];
                    bx = 3*lastX-6*f[0]+3*f[2];
                    cx = -3*lastX+3*f[0];
                    dx = lastX;
                    
                    ay = -lastY+3*f[1]-3*f[3]+f[5];
                    by = 3*lastY-6*f[1]+3*f[3];
                    cy = -3*lastY+3*f[1];
                    dy = lastY;
                    
                    //x = a*t*t*t+b*t*t+c*t+d
                    //dx/dt = 3*a*t*t+2*b*t+c
                    //t = [-B+-sqrt(B^2-4*A*C)]/(2A)
                    //A = 3*a
                    //B = 2*b
                    //C = c
                    //t = (-2*b+-sqrt(4*b*b-12*a*c)]/(6*a)
                    det = (4*bx*bx-12*ax*cx);
                    if(det<0) {
                        //there are no solutions!  nothing to do here
                    } else if(det==0) {
                        //there is 1 solution
                        t = -2*bx/(6*ax);
                        if(t>0 && t<1) {
                            x = ((ax*t+bx)*t+cx)*t+dx;
                            if(x<bounds[0]) bounds[0] = x;
                            if(x>bounds[2]) bounds[2] = x;
                        }
                    } else {
                        //there are 2 solutions:
                        det = Math.sqrt(det);
                        t = (-2*bx+det)/(6*ax);
                        if(t>0 && t<1) {
                            x = ((ax*t+bx)*t+cx)*t+dx;
                            if(x<bounds[0]) bounds[0] = x;
                            if(x>bounds[2]) bounds[2] = x;
                        }
                        
                        t = (-2*bx-det)/(6*ax);
                        if(t>0 && t<1) {
                            x = ((ax*t+bx)*t+cx)*t+dx;
                            if(x<bounds[0]) bounds[0] = x;
                            if(x>bounds[2]) bounds[2] = x;
                        }
                    }
                    
                    det = (4*by*by-12*ay*cy);
                    if(det<0) {
                        //there are no solutions!  nothing to do here
                    } else if(det==0) {
                        //there is 1 solution
                        t = -2*bx/(6*ay);
                        if(t>0 && t<1) {
                            y = ((ay*t+by)*t+cy)*t+dy;
                            if(y<bounds[1]) bounds[1] = y;
                            if(y>bounds[3]) bounds[3] = y;
                        }
                    } else {
                        //there are 2 solutions:
                        det = Math.sqrt(det);
                        t = (-2*by+det)/(6*ay);
                        if(t>0 && t<1) {
                            y = ((ay*t+by)*t+cy)*t+dy;
                            if(y<bounds[1]) bounds[1] = y;
                            if(y>bounds[3]) bounds[3] = y;
                        }
                        
                        t = (-2*by-det)/(6*ay);
                        if(t>0 && t<1) {
                            y = ((ay*t+by)*t+cy)*t+dy;
                            if(y<bounds[1]) bounds[1] = y;
                            if(y>bounds[3]) bounds[3] = y;
                        }
                    }
                    
                    lastX = f[4];
                    lastY = f[5];
                }
            }
            i.next();
        }
        
        if(defined==false) {
            throw new IllegalArgumentException("Empty shape.");
        }
        dest.setFrame(bounds[0],bounds[1],bounds[2]-bounds[0],bounds[3]-bounds[1]);
       
        return dest;
    }
}
